/*  commandir_send  Version 0.3
    Initial Release: Feb 9, 2011  
    0.2 Release: March 26, 2011
    0.3 Release: June 16, 2011
 */

#include "cmdirutils_version.h"
#include "libcmdir.h"
#include "libtcp_client.h"
#include "libcmdir_send.h"

extern unsigned int tx_bitmask;
extern char * send_filename;

unsigned char standalone = 1;

int main(int argc, char **argv)
{
  char cmdLine[16384];
  int c, baseCmdLineLen, allow_raw;
  int x, emitter_num, y;
  int rawCodes = 0;


  if(argc == 1)
  {
    printf("commandir_send %s\n\n", utils_version);
    printf(" Send IR codes to CommandIR (or commandird) for transmitting\n\n");
    printf("commandir_send [-n] [-dDELAY] [[-bEMITTER_BITMASK_HEX] | [-eEMITTERNUM] -sCOMMANDIR_NAME] FILENAME [FILENAME]...\n");
    printf("commandir_send [-n] [-dDELAY] [[-bEMITTER_BITMASK_HEX] | [-eEMITTERNUM] -sCOMMANDIR_NAME] -r [RAW CODE STRING]\n\n");
    printf(" -n[HostName] - Use a network connection to send the command to a commandird\n");
    printf("                server.  Defaults to localhost if no host specified. \n");
    printf(" -s used in conjunction with commandir_set and CommandIR Firmware 3.42+\n");
    printf("Examples:\n Send two commands to emitter 3 with a 500ms delay in between:\n");
    printf("  commandir_send -d500 -e3 amp_volup amp_volup\n");
    printf(" Send one command to emitters 1, 2, 3 and 4:\n commandir_send -b0F sa_power\n");
    printf(" Send two commands to emitter 2 on a specific CommandIR named 'COMMANDIR_0':\n");
    printf("  commandir_send -sCOMMANDIR_0 -e2 amp_volup amp_volup\n");
    printf(" Send two commands to emitter 2 on a remote CommandIR using commandird:\n");
    printf("  commandir_send -n192.168.1.15 -e2 amp_volup amp_volup\n");
    printf(" Send command Hex Code directly:\n");
    printf("  commandir_send -bff -r 0000 0074 0017 0000 0020 0021 0020 0020 0020 0021 0020 0020 0020 0021 0020 0020 0020 0021 0020 0020 0040 0020 0020 0041 003F 0041 0020\n\n");
    return 0;
  }

  set_all_bitmask(0xff);
  
  /* Scan for network command, then process args accordingly */
  for(x=1;x<argc;x++)
  {
    if((argv[x][0] == '-') && (argv[x][1] == 'n'))
    {
      // Network send - default to localhost if none specified
      networkSend = 1;
      if(strlen(argv[x]) > 2) {
        strncpy(networkSendName, &argv[x][2], strlen(argv[x]));
        networkSendName[strlen(argv[x]) - 2] = '\0';
        networkSendName[strlen(argv[x])-1] = '\0';
      } else {
        strncpy(networkSendName, localhostPtr, strlen(localhostPtr));
      }
    }
    if((argv[x][0] == '-') && (argv[x][1] == 'r'))
    {
      if(argc > x)
        rawCodes = x + 1;
      else {
        printf("Raw code required for -r switch\n");
        exit(-1);
      }
    
    }
  }  
  
  
  // -r or file names specified?  -r gets passed directly, file names converted
  if(networkSend == 1) {
    // Direct network send
    if(!networkClientConnect(networkSendName)) {
      printf("Unable to connect to %s\n", networkSendName);
      return -1;
    }

    /* Make this easy: Pass the whole command line to the commandird server */
    c = 15; // strlen("commandir_send ");
    strncpy(cmdLine, "commandir_send ", c);
    allow_raw = 0;
    for(x=1;x<argc;x++) {
      if( (argv[x][0]!='-') && (allow_raw == 0) )
        continue; // do not send file names
      if(argv[x][0]=='-' && argv[x][1]=='r')
        allow_raw = 1;
      if(argv[x][0]=='-' && argv[x][1]=='n')
        continue; // do not send the Network send flag
      if(x==argc-1)
	      c+= sprintf(&cmdLine[c], "%s", argv[x]);	// No space on final arg.
      else
	      c+= sprintf(&cmdLine[c], "%s ", argv[x]);
    }
    if(rawCodes > 0) {
      /* Nothing else to add to Raw stream, except \0 */
      cmdLine[c++] = '\0';
      printf("Sending '%s' to commandird\n", cmdLine);
      networkClientSend(cmdLine, c);
    } else {
      c+= sprintf(&cmdLine[c], "-r ");
      baseCmdLineLen = c;   // reusing the same base a few times
      // Load the file hex, and send each as a commandir_send -r command
      for(x=1;x<argc;x++) {
        if(argv[x][0]!='-')
        {
          /* Load a hex file to transmit to commandird */
          c = baseCmdLineLen;
          c += sprintf(&cmdLine[baseCmdLineLen], "%s", load_file(argv[x]));
          cmdLine[c++] = '\0';  // replace the trailing space with a \0
//          cmdLine[c-1] = '\0';  // replace the trailing space with a \0
          printf("Sending this command (%d bytes): '%s'\n", c, cmdLine);
          networkClientSend(cmdLine, c);
          printf("Sent '%s' to commandird\n", argv[x]);
        }
      }
    }
    networkClientClose();
    return 0;
  }
  
  // Stand-alone local command:
  execute_commandir_send_arg(argc, argv);
  
  return 0;
}

