/* commandir_timing
  New in Utils 0.62 - Show raw timing, from External Receiver OR Recorder
  
 */

#include "cmdirutils_version.h"
#include "libcmdir.h"
#include "libtcp_client.h"

extern unsigned char noHeader;
unsigned char standalone = STANDALONE_TIMING;
//unsigned char networkSend = 0;
//char * networkSendName;
void displayHelp(void);
void waitForTcpRecordData(void);
int execute_commandir_timing_arg(int argc, char *argv[]);

void waitForTcpRecordData() 
{
  unsigned char * dat;
  int displayed;
  int len;
  int c;
  
  dat = malloc(16384);
  
  len = networkClientRecv(dat, 16384);
  printf("IR codes from network CommandIR (press <enter> or <Ctrl+C> to quit):\n");
  
  while(1)
  {
    c = kbhit();
    if(c) {
      c = getchar();
      while(kbhit())  // clear the keyboard buffer, incase they hit keys other than enter
        c = getchar();
      break;
    }
    // len should be returned when we get a terminating \0 so we know we have a full HexString
    // Note that the network might return multiple signals!
    if(len > 0) {
      dat[len] = '\0';  // Zero terminate it at the correct place!
      displayed = 0;
      while(displayed < len) {
        displayAndPipeSignal(&dat[displayed], strlen(&dat[displayed]), 0);  
        displayed += strlen(&dat[displayed]) + 1;  // +1?
      }
    }
#ifdef WIN
    Sleep(5);
#else
    usleep(5000);
#endif
    len = networkClientRecv(dat, 16384);
//    printf("Received %d bytes: (%p) %s\n", len, dat, dat);
  }
  // When it's done, return and close
}

void displayHelp() 
{
  printf("commandir_timing %s\n\n", utils_version);
  printf(" Stream Raw IR timing codes to/from a CommandIR III\n\n");

  printf("Usage: commandir_timing [-n] [-sCOMMANDIR_NAME] [-w] [-h]\n");
  printf(" -w - Display codes without adding line-breaks or header.\n");
  printf(" -sCOMMANDIR_NAME - Record from a specific CommandIR (Firmware >=3.42 required)\n");
  printf(" -n[HostName] - Receive from a CommandIR via the 'commandird' daemon.\n");
  printf("      Defaults to localhost daemon if no host is specified. \n");
  printf(" -h - Display this help information.\n");
}

int main(int argc, char **argv)
{
  int x, y, c;
  struct commandir_device * pcd = NULL;
  char * cmdLine;
  
  // Is this a network command, or a stand-alone?
  // commandird only cares about -s and -g
  // Only for us: -d -p -n
  for(x=1;x<argc;x++)
  {
    if(argv[x][0] == '-')
    {
      switch(argv[x][1])
      {
        case 'h':
          displayHelp();
          return 0;
        case 'n':
        
          networkSend = 1;
          if(strlen(argv[x]) > 2) {
//            networkSendName = malloc(strlen(argv[x])-1); // still 1 extra byte
            strncpy(networkSendName, &argv[x][2], strlen(argv[x]));
            networkSendName[strlen(argv[x])] = '\0';
          } else {
            strncpy(networkSendName, localhostPtr, strlen(localhostPtr));
//            networkSendName = localhostPtr;
          }

          /* CONNECT */
          if(!networkClientConnect(networkSendName)) {
            printf("Unable to connect to %s\n", networkSendName);
            return -1;
          }
          
          /* Make this easy: Pass the whole command line to the commandird server */
          cmdLine = malloc(16384);
          c = strlen("commandir_timing ");
          strncpy(cmdLine, "commandir_timing ", c);
          for(y=1;y<argc;y++) {
            if(argv[y][0]=='-' && argv[y][1]=='n')
              continue; // do not send the Network send flag
            c+= sprintf(&cmdLine[c], "%s ", argv[y]);
          }
          cmdLine[c++] = '\0';
          networkClientSend(cmdLine, c);
          break;
        case 's':
          break;
        case 'w':
          noHeader = 1;
          break;
      }
    }
  } // end arg for
  
  if(networkSend == 1) 
  {
    /* This case is for continuous operation - display or pipe only 
       commandird must send to us in timing format.
     */
    waitForTcpRecordData();
    networkClientClose();
    return 0;  
  } else {
    printf("commandir_timing %s\n\n", utils_version);
  }

    // Stand-alone local command:
  return execute_commandir_timing_arg(argc, argv);
  
} // end main

